/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2009 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * Original version of this file contributed by Mikko Lehtonen.
 *
 * $Id: T2BMFont.cpp 144 2009-08-17 17:02:29Z jlauha $
 * $Revision: 144 $
 */

#include <cstdio>
#include <cstring>
#include <cassert>

#include "Turs2.h"

#ifdef _MSC_VER
#ifndef snprintf
#define snprintf _snprintf
#endif
#endif // _MSC_VER

using std::memset;
using std::strncpy;
using std::fopen;
using std::fscanf;

namespace turska
{

static C8 sGetEscapedChar(C8 successor)
{
    switch (successor)    // todo: wider support for different escapes
    {
    case 'n': return '\n';
    case 'r': return '\r';
    case 't': return '\t';
    }
    return successor; // unknown
}


BMFont::FontData::FontData() : pages(0), kernings(0), chars(0)
{
}

BMFont::FontData::~FontData()
{
    reset();
}

void BMFont::FontData::reset()
{
    delete[] pages;
    delete[] kernings;
    delete[] chars;
    pages = 0;
    kernings = 0;
    chars = 0;
}

BMFont::Page::Page() : fontImage(0)
{
}

BMFont::Page::~Page()
{
    delete fontImage;
}


BMFont::BMFont(Graphics *graphics, const C8 *descriptionFile) :
    mSpacing(0), mGraphics(graphics)
{
    memset(mFontData.info.face, 0, MAX_FIELD_LENGTH);
    memset(mFontData.info.charset, 0, MAX_FIELD_LENGTH);
    memset(mFileName, 0, MAX_FILENAME_LENGTH);
    strncpy(mFileName, descriptionFile, MAX_FILENAME_LENGTH - 1);
}

BMFont::~BMFont()
{
}


bool BMFont::init()
{
    return reload();
}


bool BMFont::reload()
{
    I32 a;

    mLoaded = false;

    mFontData.reset();

    FILE *fp = fopen(mFileName, "r");

    if (fp == 0)
        return false;

    /* Note: fails in following cases:
     * - if charset includes whitespace
     * - if face name is longer than max field length-1
     * - if charset is longer than max field length-2
     */
    // keep the %NUM as (MAX_FIELD_LENGTH-1)
    I32 size, stretchH, padding[4], spacing[2];
    I32 outline, aa, bold, italic, unicode, smooth;
    if (fscanf(fp, "info face=\"%63[^\"]\" size=%d bold=%d italic=%d charset=\"%63s unicode=%d stretchH=%d smooth=%d aa=%d padding=%d,%d,%d,%d spacing=%d,%d outline=%d%*[\n\r]",
               mFontData.info.face,
               &size,
               &bold,
               &italic,
               mFontData.info.charset,
               &unicode,
               &stretchH,
               &smooth,
               &aa,
               &padding[0],
               &padding[1],
               &padding[2],
               &padding[3],
               &spacing[0],
               &spacing[1],
               &outline) != 16)
    {
        fclose(fp);
        return false;
    }
    mFontData.info.size = (UI16)size;
    mFontData.info.bold = bold ? true : false;
    mFontData.info.italic = italic ? true : false;
    mFontData.info.unicode = unicode ? true : false;
    mFontData.info.stretchH = (I16)stretchH;
    mFontData.info.smooth = smooth ? true : false;
    mFontData.info.aa = (I8)aa;
    mFontData.info.padding[PADDING_UP] = (I16)padding[0];
    mFontData.info.padding[PADDING_RIGHT] = (I16)padding[1];
    mFontData.info.padding[PADDING_DOWN] = (I16)padding[2];
    mFontData.info.padding[PADDING_LEFT] = (I16)padding[3];
    mFontData.info.horizSpacing = (I16)spacing[0];
    mFontData.info.vertSpacing = (I16)spacing[1];
    mFontData.info.outline = (I16)outline;

    // Charset is handled differently as it can be zero size string
    mFontData.info.charset[strlen(mFontData.info.charset) - 1] = 0; // delete ending quote

    I32 lineHeight, base, scaleW, scaleH, pages, packed;
    I32 alphaChnl, redChnl, greenChnl, blueChnl;
    if (fscanf(fp, "common lineHeight=%d base=%d scaleW=%d scaleH=%d pages=%d packed=%d alphaChnl=%d redChnl=%d greenChnl=%d blueChnl=%d%*[\n\r]",
               &lineHeight,
               &base,
               &scaleW,
               &scaleH,
               &pages,
               &packed,
               &alphaChnl,
               &redChnl,
               &greenChnl,
               &blueChnl) != 10)
    {
        fclose(fp);
        return false;
    }

    mFontData.common.lineHeight = (UI16)lineHeight;
    mFontData.common.base = (I16)base;
    mFontData.common.scaleW = (I16)scaleW;
    mFontData.common.scaleH = (I16)scaleH;
    mFontData.common.pages = (UI16)pages;
    mFontData.common.packed = packed ? true : false;
    mFontData.common.alphaChnl = (UI8)alphaChnl;
    mFontData.common.redChnl = (UI8)redChnl;
    mFontData.common.greenChnl = (UI8)greenChnl;
    mFontData.common.blueChnl = (UI8)blueChnl;

    mFontData.pages = new Page[mFontData.common.pages];
    if (!mFontData.pages)
    {
        fclose(fp);
        return false;
    }

    for (a = 0; a < mFontData.common.pages; ++a)
    {
        I32 id;
        // Note: keep the %NUM as (MAX_FILENAME_LENGTH-1)
        if (fscanf(fp, "page id=%d file=\"%127[^\"]\"%*[\n\r]",
                   &id,
                   mFontData.pages[a].file) != 2)
        {
            fclose(fp);
            return false;
        }
        mFontData.pages[a].id = (UI16)id;
    }

    if (fscanf(fp, "chars count=%d\n", &mFontData.charCount) != 1)
    {
        fclose(fp);
        return false;
    }

    bool foundSpaceChar = false;
    int spacePos = 0;
    // allocate one extra for the potential explicitly added space
    mFontData.chars = new Char[mFontData.charCount + 1];

    for (a = 0; a < mFontData.charCount; ++a)
    {
        I32 x, y, width, height, xoffset, yoffset;
        I32 xadvance, page, chnl;
        if (fscanf(fp, "char%*[ ]id=%d%*[ ]x=%d%*[ ]y=%d%*[ ]width=%d%*[ ]height=%d%*[ ]xoffset=%d%*[ ]yoffset=%d%*[ ]xadvance=%d%*[ ]page=%d%*[ ]chnl=%d%*[\n\r]",
                   &mFontData.chars[a].id,
                   &x,
                   &y,
                   &width,
                   &height,
                   &xoffset,
                   &yoffset,
                   &xadvance,
                   &page,
                   &chnl) != 10)
        {
            fclose(fp);
            return false;
        }
        mFontData.chars[a].x = (UI16)x;
        mFontData.chars[a].y = (UI16)y;
        mFontData.chars[a].width = (UI16)width;
        mFontData.chars[a].height = (UI16)height;
        mFontData.chars[a].xoffset = (I16)xoffset;
        mFontData.chars[a].yoffset = (I16)yoffset;
        mFontData.chars[a].xadvance = (I16)xadvance;
        mFontData.chars[a].page = (UI16)page;
        mFontData.chars[a].chnl = (UI8)chnl;
        mFontData.chars[a].firstKernIndex = -1;

        if (mFontData.chars[a].id == 32)
            foundSpaceChar = true;
        if (mFontData.chars[a].id < 32)
            ++spacePos;
    }

    if (!foundSpaceChar)
    {
        I32 charsToMove = mFontData.charCount - spacePos;
        if (charsToMove > 0)
            memmove(&mFontData.chars[spacePos + 1],
                    &mFontData.chars[spacePos],
                    charsToMove * sizeof(Char));
        Char &space = mFontData.chars[spacePos];
        space.id = 32;
        space.x = space.y = 0;
        space.width = space.height = 0;
        space.xoffset = space.yoffset = 0;
        // default space width is third of font size
        space.xadvance = mFontData.info.size / 3;
        space.page = 0;
        space.chnl = 0;
        ++mFontData.charCount;
    }

    if (fscanf(fp, "kernings count=%d%*[\n\r]", &mFontData.kerningCount) != 1)
        mFontData.kerningCount = 0;

    if (mFontData.kerningCount > 0)
    {
        mFontData.kernings = new Kerning[mFontData.kerningCount];
        if (!mFontData.kernings)
        {
            fclose(fp);
            return false;
        }

        for (a = 0; a < mFontData.kerningCount; ++a)
        {
            I32 amount;
            if (fscanf(fp, "kerning first=%d%*[ ]second=%d%*[ ]amount=%d%*[ \n\r]",
                       &mFontData.kernings[a].first,
                       &mFontData.kernings[a].second,
                       &amount) != 3)
            {
                fclose(fp);
                return false;
            }
            mFontData.kernings[a].amount = (I16)amount;
            I32 idx = findChar(mFontData.kernings[a].first);
            if (idx >= 0 && mFontData.chars[idx].firstKernIndex == -1)
                mFontData.chars[idx].firstKernIndex = a;
        }
    } // kerningCount > 0

    fclose(fp);

    // Try to load images
    for (a = 0; a < mFontData.common.pages; ++a)
    {
        mFontData.pages[a].fontImage = mGraphics->loadImage(mFontData.pages[0].file);
        if (mFontData.pages[a].fontImage == 0)
        {
            // failed - try to load it relatively at the same path as the description file
            C8 dir[128];
            C8 *last = strrchr(mFileName, '/');
            if (last == 0)
                last = strrchr(mFileName, '\\');
            I32 len = (I32)(last - mFileName) + 1;
            if (last != 0 && len < 128)
            {
                strncpy(dir, mFileName, len);
                dir[len] = 0;
                C8 fullpath[256] = { 0 };
                snprintf(fullpath, 255, "%s%s", dir, mFontData.pages[a].file);
                mFontData.pages[a].fontImage = mGraphics->loadImage(fullpath);
                if (mFontData.pages[a].fontImage == 0)
                    return false;
            }
        }
        assert( mFontData.pages[a].fontImage );
    }
    
    mLoaded = true;

    return true;
}


I32 BMFont::findChar(UI32 ch)
{
    // Performs a binary search.
    // Assumes that the chars array is sorted.
    I32 l = 0;
    I32 r = mFontData.charCount - 1;
    while (l <= r)
    {
        I32 m = (l + r) / 2;
        const UI32 id = mFontData.chars[m].id;
        if (id > ch)
            r = m - 1;
        else if (id < ch)
            l = m + 1;
        else
            return m;
    }
    return -1;
}


I32 BMFont::findKern(I32 first, I32 second)
{
    return findKern(first, second, findChar(first));
}

I32 BMFont::findKern(I32 first, I32 second, I32 firstCharIndex)
{
    I32 a, start = 0;
    if (firstCharIndex >= 0)
    {
        start = mFontData.chars[firstCharIndex].firstKernIndex;
        if (start < 0)
            return 0;
    }
    const I32 kerningCount = mFontData.kerningCount;
    for (a = start; a < kerningCount; ++a)
    {
        if (mFontData.kernings[a].first == first)
        {
            if (mFontData.kernings[a].second == second)
                return mFontData.kernings[a].amount;
        }
        else
        {
            if (a > 0 && mFontData.kernings[a - 1].first == first)
                break; // reached end of candidates
        }
    }
    return 0;
}


const C8 * BMFont::getFontName() const
{
    return mFontData.info.face;
}


I32 BMFont::getFontHeight()
{
    return mFontData.common.lineHeight;
}


TREAL BMFont::getFontHeight(TREAL scale)
{
    return mFontData.common.lineHeight * scale;
}


bool BMFont::hasCharacter(UI32 character)
{
    return findChar(character) >= 0;
}


TREAL BMFont::getStringWidth(const C8 *string, TREAL scale, I32 substrLength)
{
    const I32 maxLength = substrLength >= 0 ? substrLength : 0x7fffffff;
    I32 a, w = 0, last = -1;
    bool escape = false;
    for (a = 0; a < maxLength && string[a]; ++a)
    {
        C8 currentChar = string[a];
        if (getFlags() & Font::FF_HANDLE_BACKSLASH_ESCAPES)
        {
            if (escape)
            {
                currentChar = sGetEscapedChar(currentChar);
                escape = false;
            }
            else if (currentChar == '\\')
            {
                escape = true;
                continue;
            }
        }

        UI32 ch = currentChar & 0xff;
        I32 id = findChar(ch);
        if (id == -1)
        {
            last = -1;
            continue;
        }
        if (last != -1)
            w += findKern(last, ch);
        w += mFontData.chars[id].xadvance + mSpacing;
        last = ch;
    }
    // subtract the spacing after last character from result, scale final result
    return (w - mSpacing) * scale;
}


TREAL BMFont::drawString(TREAL x, TREAL y, const C8 *string,
                        UI32 flags, TREAL scale,
                        const Color &color, I32 substrLength)
{
    I32 a, last = -1, lastIdx = -1, maxLength = (substrLength >= 0) ? substrLength : 0x7fffffff;
    Rectangle rect;
    TREAL xoff = x;
    TREAL yoff = y;
    TREAL stringWidth = 0;

    if ((flags & IF_ALIGN_HORIZ_CENTER) != 0)
        xoff -= getStringWidth(string, scale, substrLength) / TREAL(2);
    else if ((flags & IF_ALIGN_RIGHT) != 0)
        xoff -= getStringWidth(string, scale, substrLength);
    if ((flags & IF_ALIGN_VERT_CENTER) != 0)
        yoff -= getFontHeight(scale) / TREAL(2);
    else if ((flags & IF_ALIGN_BOTTOM) != 0)
        yoff -= getFontHeight(scale);

    if (scale == TREAL(1))
    {
        x = (TREAL)((I32)x);
        y = (TREAL)((I32)y);
    }
    for (a = 0; a < mFontData.common.pages; ++a)
    {
        assert(mFontData.pages[a].fontImage);
        mFontData.pages[a].fontImage->setScale(scale);
    }

    flags = IF_ALIGN_TOP_LEFT | IF_COLOR | IF_ALPHA;
    if (scale != (TREAL)1)
        flags |= IF_TRANSFORM | IF_SCALE;

    bool escape = false;

    for (a = 0; a < maxLength && string[a]; ++a)
    {
        C8 currentChar = string[a];
        if (getFlags() & Font::FF_HANDLE_BACKSLASH_ESCAPES)
        {
            if (escape)
            {
                currentChar = sGetEscapedChar(currentChar);
                escape = false;
            }
            else if (currentChar == '\\')
            {
                escape = true;
                continue;
            }
        }

        UI32 ch = currentChar & 0xff;
        I32 id = findChar(ch);

        if (last != -1)
            xoff += findKern(last, ch, lastIdx) * scale;
        if (id == -1)
        {
            last = -1;
            continue;
        }
        rect.x = static_cast<TREAL>(mFontData.chars[id].x);
        rect.y = static_cast<TREAL>(mFontData.chars[id].y);
        rect.width = static_cast<TREAL>(mFontData.chars[id].width);
        rect.height = static_cast<TREAL>(mFontData.chars[id].height);

        TREAL x, y;
        TREAL gw = TREAL(mFontData.chars[id].xadvance + mSpacing);
        if (scale != (TREAL)1)
        {
            x = xoff + mFontData.chars[id].xoffset * scale;
            y = yoff + mFontData.chars[id].yoffset * scale;
            gw *= scale;
        }
        else
        {
            x = (TREAL)((I32)(xoff + mFontData.chars[id].xoffset));
            y = (TREAL)((I32)(yoff + mFontData.chars[id].yoffset));
        }
        stringWidth += gw;
        xoff += gw;
        I32 page = mFontData.chars[id].page;
        mFontData.pages[page].fontImage->draw(mDrawMode, flags, x, y, color, &rect);

        last = ch;
        lastIdx = id;
    } // a < len

    return stringWidth - mSpacing * scale;
}


} // namespace turska
