/* Turska Framework/UI library (Turs2)
 * Copyright 2003-2009 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/turska/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: T2ImageOpenGL.cpp 2393 2009-08-14 10:38:46Z tonic $
 * $Revision: 2393 $
 */

#include <math.h>
#include "Turs2.h"

#ifdef T2_USE_SDLOPENGL

// SDL is included here only for the OpenGL.
#include "SDL.h"
#include "SDL_opengl.h"


namespace turska
{


ImageOpenGL::ImageOpenGL(UI32 width, UI32 height, UI32 flags,
                         UI32 glId, const F32 *uv) :
    Image(width, height, IDM_NORMAL, flags),
    mGLId(glId),
    mHalfWidth((F32)width / 2.0f),
    mHalfHeight((F32)height / 2.0f),
    mTargetWidth((F32)width),
    mTargetHeight((F32)height),
    mAngle(0),
    mScaleX(1), mScaleY(1),
    mUsingExplicitTargetSize(false)
{
    UI32 a;
    for (a = 0; a < 4; ++a)
        mUV[a] = uv[a];
    mColor.set(0xffffffff);
}


ImageOpenGL::~ImageOpenGL()
{
    glDeleteTextures(1, (const GLuint *)&mGLId);
}


void ImageOpenGL::draw(IMAGE_DRAW_MODE drawMode, UI32 flags, F32 x, F32 y, const UI8 *color, const Rectangle *sourceRect)
{
    bool usingSrcArea = true;
    Rectangle srcArea;
    if (sourceRect)
        srcArea = *sourceRect;
    else if (!sourceRect && mUsingSourceRectangle)
        srcArea = mSourceRectangle;
    else
    {
        srcArea = Rectangle(0, 0, (TREAL)mWidth, (TREAL)mHeight);
        usingSrcArea = false;
    }

    // other than normal drawMode needs storing of attributes
    const bool storeAttrib = (drawMode != IDM_NORMAL);

    if (storeAttrib)
        glPushAttrib(GL_COLOR_BUFFER_BIT);

    switch (drawMode)
    {
    case IDM_ADD:
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        break;
    case IDM_MULTIPLY:
        glBlendFunc(GL_ZERO, GL_SRC_COLOR);
        break;
    case IDM_PREMULT:
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        break;
    default:;
        // blend func is assumed to be (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) by default
    }

    UI8 finalColor[4];
    if (flags & IF_COLOR)
    {
        for (I32 a = 0; a < 3; ++a)
            finalColor[a] = color[a];
    }
    else
    {
        for (I32 a = 0; a < 3; ++a)
            finalColor[a] = (UI8)0xff;
    }
    if (flags & IF_ALPHA)
        finalColor[3] = color[3];
    else
        finalColor[3] = (UI8)0xff;
    glColor4ubv(finalColor);

    glBindTexture(GL_TEXTURE_2D, mGLId);

    F32 uv[4];
    memcpy(uv, mUV, 4 * sizeof(F32));
    if (flags & IF_FLIP_HORIZ)
    {
        uv[0] = mUV[2];
        uv[2] = mUV[0];
    }
    if (flags & IF_FLIP_VERT)
    {
        uv[1] = mUV[3];
        uv[3] = mUV[1];
    }
    F32 halfWidth = mHalfWidth, halfHeight = mHalfHeight;
    if (usingSrcArea)
    {
        F32 uvW = uv[2] - uv[0];
        F32 uvH = uv[3] - uv[1];
        uv[0] += uvW * srcArea.x / mWidth;
        uv[1] += uvH * srcArea.y / mHeight;
        uv[2] = uv[0] + uvW * srcArea.width / mWidth;
        uv[3] = uv[1] + uvH * srcArea.height / mHeight;
        halfWidth = srcArea.width * 0.5f;
        halfHeight = srcArea.height * 0.5f;
    }
    
    if (flags & IF_TRANSFORM)
    {
        F32 pivotX = -halfWidth, pivotY = -halfHeight; // transform center adjusted to top-left (default)
        if (flags & IF_ALIGN_RIGHT)
            pivotX = halfWidth;
        if (flags & IF_ALIGN_HORIZ_CENTER)
            pivotX = (mUsingCustomPivot ? mPivotX : halfWidth) - halfWidth;
        if (flags & IF_ALIGN_BOTTOM)
            pivotY = halfHeight;
        if (flags & IF_ALIGN_VERT_CENTER)
            pivotY = (mUsingCustomPivot ? mPivotY : halfHeight) - halfHeight;

        F32 vx1[4], vy1[4], vx2[4], vy2[4];
        F32 *vx = vx1, *vy = vy1;
        vx[0] = -halfWidth - pivotX; vy[0] = -halfHeight - pivotY;
        vx[1] = halfWidth - pivotX; vy[1] = -halfHeight - pivotY;
        vx[2] = -halfWidth - pivotX; vy[2] = halfHeight - pivotY;
        vx[3] = halfWidth - pivotX; vy[3] = halfHeight - pivotY;
        if (flags & IF_SCALE)
        {
            vx[0] *= mScaleX; vy[0] *= mScaleY;
            vx[1] *= mScaleX; vy[1] *= mScaleY;
            vx[2] *= mScaleX; vy[2] *= mScaleY;
            vx[3] *= mScaleX; vy[3] *= mScaleY;
        }
        if (flags & IF_ROTATE)
        {
            F32 c = cosf(mAngle), s = sinf(mAngle);
            // rotate current vertices to temp vertices
            vx2[0] = vx[0] * c - vy[0] * s; vy2[0] = vx[0] * s + vy[0] * c;
            vx2[1] = vx[1] * c - vy[1] * s; vy2[1] = vx[1] * s + vy[1] * c;
            vx2[2] = vx[2] * c - vy[2] * s; vy2[2] = vx[2] * s + vy[2] * c;
            vx2[3] = vx[3] * c - vy[3] * s; vy2[3] = vx[3] * s + vy[3] * c;
            // switch temp vertices to final ones
            vx = vx2; vy = vy2;
        }

        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(uv[0], uv[1]); glVertex2f(vx[0] + x, vy[0] + y);
        glTexCoord2f(uv[2], uv[1]); glVertex2f(vx[1] + x, vy[1] + y);
        glTexCoord2f(uv[0], uv[3]); glVertex2f(vx[2] + x, vy[2] + y);
        glTexCoord2f(uv[2], uv[3]); glVertex2f(vx[3] + x, vy[3] + y);
        glEnd();
    }
    else
    {
        F32 targetWidth = mTargetWidth, targetHeight = mTargetHeight;
        if (usingSrcArea && !mUsingExplicitTargetSize)
        {
            targetWidth = srcArea.width;
            targetHeight = srcArea.height;
        }
        F32 topLeftX = x, topLeftY = y; // top-left (default)
        if (flags & IF_ALIGN_RIGHT)
            topLeftX -= targetWidth;
        if (flags & IF_ALIGN_HORIZ_CENTER)
        {
            const F32 normalized = ((mUsingCustomPivot ? mPivotX : halfWidth) - halfWidth) / halfWidth; // range -1..1
            const F32 halfTargetWidth = targetWidth * 0.5f;
            topLeftX -= halfTargetWidth + halfTargetWidth * normalized;
        }
        if (flags & IF_ALIGN_BOTTOM)
            topLeftY -= targetHeight;
        if (flags & IF_ALIGN_VERT_CENTER)
        {
            const F32 normalized = ((mUsingCustomPivot ? mPivotY : halfHeight) - halfHeight) / halfHeight; // range -1..1
            const F32 halfTargetHeight = targetHeight * 0.5f;
            topLeftY -= halfTargetHeight + halfTargetHeight * normalized;
        }

        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(uv[0], uv[1]); glVertex2f(topLeftX, topLeftY);
        glTexCoord2f(uv[2], uv[1]); glVertex2f(topLeftX + targetWidth, topLeftY);
        glTexCoord2f(uv[0], uv[3]); glVertex2f(topLeftX, topLeftY + targetHeight);
        glTexCoord2f(uv[2], uv[3]); glVertex2f(topLeftX + targetWidth, topLeftY + targetHeight);
        glEnd();
    }

    if (storeAttrib)
        glPopAttrib();
}


} // namespace turska
#endif // T2_USE_SDLOPENGL
